@@ -463,6 +463,13 @@ xfs_dummy_verify(
 	return;
 }
 
+void
+xfs_verify_recalc_crc(
+	struct xfs_buf *bp)
+{
+	xfs_buf_update_cksum(bp, iocur_top->typ->crc_off);
+}
+
 void
 write_cur(void)
 {
