@@ -150,6 +150,12 @@ extern void		fiemap_init(void);
 #define fiemap_init()	do { } while (0)
 #endif
 
+#ifdef HAVE_COPY_FILE_RANGE
+extern void		copy_range_init(void);
+#else
+#define copy_range_init()	do { } while (0)
+#endif
+
 #ifdef HAVE_SYNC_FILE_RANGE
 extern void		sync_range_init(void);
 #else
@@ -161,3 +167,5 @@ extern void		readdir_init(void);
 #else
 #define readdir_init()		do { } while (0)
 #endif
+
+extern void		reflink_init(void);
