@@ -16,18 +16,17 @@
  * Inc.,  51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
-#include <libxfs.h>
-#include <libxlog.h>
+#include "libxfs.h"
+#include "libxcmd.h"
+#include "libxlog.h"
 #include "agheader.h"
 #include "globals.h"
 #include "protos.h"
 #include "err_protos.h"
+#include "xfs_multidisk.h"
 
 #define BSIZE	(1024 * 1024)
 
-#define XFS_AG_BYTES(bblog)	((long long)BBSIZE << (bblog))
-#define	XFS_AG_MIN_BYTES	((XFS_AG_BYTES(15)))	/* 16 MB */
-
 /*
  * copy the fields of a superblock that are present in primary and
  * secondaries -- preserve fields that are different in the primary.
@@ -84,11 +83,31 @@ copy_sb(xfs_sb_t *source, xfs_sb_t *dest)
 	memset(source->sb_fname, 0, 12);
 }
 
-/*
- * find a secondary superblock, copy it into the sb buffer
- */
 int
-find_secondary_sb(xfs_sb_t *rsb)
+verify_sb_blocksize(xfs_sb_t *sb)
+{
+	/* check to make sure blocksize is legal 2^N, 9 <= N <= 16 */
+	if (sb->sb_blocksize == 0)
+		return XR_BAD_BLOCKSIZE;
+	if (sb->sb_blocksize != (1 << sb->sb_blocklog))
+		return XR_BAD_BLOCKLOG;
+	if (sb->sb_blocklog < XFS_MIN_BLOCKSIZE_LOG ||
+	    sb->sb_blocklog > XFS_MAX_BLOCKSIZE_LOG)
+		return XR_BAD_BLOCKLOG;
+
+	return 0;
+}
+
+/*
+ * find a secondary superblock, copy it into the sb buffer.
+ * start is the point to begin reading BSIZE bytes.
+ * skip contains a byte-count of how far to advance for next read.
+ */
+static int
+__find_secondary_sb(
+	xfs_sb_t	*rsb,
+	__uint64_t	start,
+	__uint64_t	skip)
 {
 	xfs_off_t	off;
 	xfs_sb_t	*sb;
@@ -100,8 +119,6 @@ find_secondary_sb(xfs_sb_t *rsb)
 	int		retval;
 	int		bsize;
 
-	do_warn(_("\nattempting to find secondary superblock...\n"));
-
 	sb = (xfs_sb_t *)memalign(libxfs_device_alignment(), BSIZE);
 	if (!sb) {
 		do_error(
@@ -117,7 +134,7 @@ find_secondary_sb(xfs_sb_t *rsb)
 	/*
 	 * skip first sector since we know that's bad
 	 */
-	for (done = 0, off = XFS_AG_MIN_BYTES; !done ; off += bsize)  {
+	for (done = 0, off = start; !done ; off += skip)  {
 		/*
 		 * read disk 1 MByte at a time.
 		 */
@@ -138,7 +155,6 @@ find_secondary_sb(xfs_sb_t *rsb)
 		for (i = 0; !done && i < bsize; i += BBSIZE)  {
 			c_bufsb = (char *)sb + i;
 			libxfs_sb_from_disk(&bufsb, (xfs_dsb_t *)c_bufsb);
-			libxfs_sb_quota_from_disk(&bufsb);
 
 			if (verify_sb(c_bufsb, &bufsb, 0) != XR_OK)
 				continue;
@@ -151,7 +167,7 @@ find_secondary_sb(xfs_sb_t *rsb)
 			 */
 			memmove(rsb, &bufsb, sizeof(xfs_sb_t));
 			rsb->sb_inprogress = 0;
-			clear_sunit = 1;
+			copied_sunit = 1;
 
 			if (verify_set_primary_sb(rsb, 0, &dirty) == XR_OK)  {
 				do_warn(
@@ -166,16 +182,80 @@ find_secondary_sb(xfs_sb_t *rsb)
 	}
 
 	free(sb);
-	return(retval);
+	return retval;
+}
+
+static int
+guess_default_geometry(
+	__uint64_t		*agsize,
+	__uint64_t		*agcount,
+	libxfs_init_t		*x)
+{
+	struct fs_topology	ft;
+	int			blocklog;
+	__uint64_t		dblocks;
+	int			multidisk;
+
+	memset(&ft, 0, sizeof(ft));
+	get_topology(x, &ft, 1);
+
+	/*
+	 * get geometry from get_topology result.
+	 * Use default block size (2^12)
+	 */
+	blocklog = 12;
+	multidisk = ft.dswidth | ft.dsunit;
+	dblocks = x->dsize >> (blocklog - BBSHIFT);
+	calc_default_ag_geometry(blocklog, dblocks, multidisk,
+				 agsize, agcount);
+
+	return blocklog;
+}
+
+int
+find_secondary_sb(xfs_sb_t *rsb)
+{
+	int		retval = 0;
+	__uint64_t	agcount;
+	__uint64_t	agsize;
+	__uint64_t	skip;
+	int		blocklog;
+
+	/*
+	 * Attempt to find secondary sb with a coarse approach,
+	 * first trying agblocks and blocksize read from sb, providing
+	 * they're sane.
+	 */
+	do_warn(_("\nattempting to find secondary superblock...\n"));
+
+	if (verify_sb_blocksize(rsb) == 0) {
+		skip = (__uint64_t)rsb->sb_agblocks * rsb->sb_blocksize;
+		if (skip >= XFS_AG_MIN_BYTES && skip <= XFS_AG_MAX_BYTES)
+			retval = __find_secondary_sb(rsb, skip, skip);
+	}
+
+        /* If that failed, retry coarse approach, using default geometry */
+        if (!retval) {
+                blocklog = guess_default_geometry(&agsize, &agcount, &x);
+                skip = agsize << blocklog;
+                retval = __find_secondary_sb(rsb, skip, skip);
+        }
+
+        /* If that failed, fall back to the brute force method */
+        if (!retval)
+                retval = __find_secondary_sb(rsb, XFS_AG_MIN_BYTES, BSIZE);
+
+	return retval;
 }
 
 /*
- * Calculate what inode alignment field ought to be
- * based on internal superblock info and determine if it is valid.
+ * Calculate what the inode alignment field ought to be based on internal
+ * superblock info and determine if it is valid.
  *
- * For v5 superblocks, the inode alignment will either match that of the
- * standard XFS_INODE_BIG_CLUSTER_SIZE, or it will be scaled based on the inode
- * size. Either value is valid in this case.
+ * For standard v5 superblocks, the inode alignment must either match
+ * XFS_INODE_BIG_CLUSTER_SIZE or a multiple based on the inode size. For v5
+ * superblocks with sparse inode chunks enabled, inode alignment must match the
+ * inode chunk size.
  *
  * Return true if the alignment is valid, false otherwise.
  */
@@ -196,10 +276,25 @@ sb_validate_ino_align(struct xfs_sb *sb)
 	if (!xfs_sb_version_hascrc(sb))
 		return false;
 
-	align *= sb->sb_inodesize / XFS_DINODE_MIN_SIZE;
+	align = (XFS_INODE_BIG_CLUSTER_SIZE *
+		 sb->sb_inodesize / XFS_DINODE_MIN_SIZE) >> sb->sb_blocklog;
 	if (align == sb->sb_inoalignmt)
 		return true;
 
+	/*
+	 * Sparse inodes requires inoalignmt to match full inode chunk size and
+	 * spino_align to match the scaled alignment (as calculated above).
+	 */
+	if (xfs_sb_version_hassparseinodes(sb)) {
+		if (align != sb->sb_spino_align)
+			return false;
+
+		align = (sb->sb_inodesize * XFS_INODES_PER_CHUNK)
+			>> sb->sb_blocklog;
+		if (align == sb->sb_inoalignmt)
+			return true;
+	}
+
 	return false;
 }
 
@@ -250,6 +345,7 @@ verify_sb(char *sb_buf, xfs_sb_t *sb, int is_primary_sb)
 {
 	__uint32_t	bsize;
 	int		i;
+	int		ret;
 
 	/* check magic number and version number */
 
@@ -291,23 +387,10 @@ verify_sb(char *sb_buf, xfs_sb_t *sb, int is_primary_sb)
 	    !xfs_verify_cksum(sb_buf, sb->sb_sectsize, XFS_SB_CRC_OFF))
 		return XR_BAD_CRC;
 
-	/* check to make sure blocksize is legal 2^N, 9 <= N <= 16 */
-	if (sb->sb_blocksize == 0)
-		return(XR_BAD_BLOCKSIZE);
-
-	bsize = 1;
-
-	for (i = 0; bsize < sb->sb_blocksize &&
-		i < sizeof(sb->sb_blocksize) * NBBY; i++)
-		bsize <<= 1;
-
-	if (i < XFS_MIN_BLOCKSIZE_LOG || i > XFS_MAX_BLOCKSIZE_LOG)
-		return(XR_BAD_BLOCKSIZE);
-
-	/* check sb blocksize field against sb blocklog field */
-
-	if (i != sb->sb_blocklog)
-		return(XR_BAD_BLOCKLOG);
+	/* check to ensure blocksize and blocklog are legal */
+	ret = verify_sb_blocksize(sb);
+	if (ret != 0)
+		return ret;
 
 	/* sanity check ag count, size fields against data size field */
 
@@ -411,61 +494,6 @@ verify_sb(char *sb_buf, xfs_sb_t *sb, int is_primary_sb)
 			return(XR_BAD_SB_WIDTH);
 	}
 
-	/*
-	 * if shared bit is set, verify that the version number is sane
-	 */
-	if (xfs_sb_version_hasshared(sb))  {
-		if (sb->sb_shared_vn > XFS_SB_MAX_SHARED_VN)
-			return(XR_BAD_SVN);
-	}
-
-	/*
-	 * mkfs's that stamped a feature bit besides the ones in the
-	 * mask below could leave garbage in the secondary superblock
-	 * sectors.  Anything stamping the shared fs bit or better into
-	 * the secondaries is ok and should generate clean secondary
-	 * superblock sectors.
-	 *
-	 * check primary and clean secondary superblocks more strictly
-	 */
-	if (is_primary_sb || sb->sb_versionnum & XR_PART_SECSB_VNMASK)  {
-		/*
-		 * return errors if shared vn or alignment fields
-		 * are set without their feature bits being set
-		 */
-		if ((!pre_65_beta && (sb->sb_versionnum & XR_PART_SECSB_VNMASK)) ||
-		    (pre_65_beta && (sb->sb_versionnum & XR_ALPHA_SECSB_VNMASK))) {
-			/*
-			 * shared version # and inode alignment fields
-			 * should be valid
-			 */
-			if (sb->sb_shared_vn && !xfs_sb_version_hasshared(sb))
-				return(XR_BAD_SVN);
-			if (sb->sb_inoalignmt && !xfs_sb_version_hasalign(sb))
-				return(XR_BAD_INO_ALIGN);
-		}
-		if ((!pre_65_beta &&
-		     (sb->sb_versionnum & XR_GOOD_SECSB_VNMASK)) ||
-		    (pre_65_beta &&
-		     (sb->sb_versionnum & XFS_SB_VERSION_DALIGNBIT)))  {
-			/*
-			 * stripe alignment values should be valid
-			 */
-			if (sb->sb_unit && !xfs_sb_version_hasdalign(sb))
-				return(XR_BAD_SB_UNIT);
-			if (sb->sb_width && !xfs_sb_version_hasdalign(sb))
-				return(XR_BAD_SB_WIDTH);
-		}
-
-#if 0
-		/*
-		 * checks involving later superblock fields get added here...
-		 */
-		if (sb->sb_versionnum & XR_GOOD_SECSB_VNMASK)  {
-		}
-#endif
-	}
-
 	return(XR_OK);
 }
 
@@ -489,7 +517,7 @@ write_primary_sb(xfs_sb_t *sbp, int size)
 		do_error(_("couldn't seek to offset 0 in filesystem\n"));
 	}
 
-	libxfs_sb_to_disk(buf, sbp, XFS_SB_ALL_BITS);
+	libxfs_sb_to_disk(buf, sbp);
 
 	if (xfs_sb_version_hascrc(sbp))
 		xfs_update_cksum((char *)buf, size, XFS_SB_CRC_OFF);
@@ -539,7 +567,6 @@ get_sb(xfs_sb_t *sbp, xfs_off_t off, int size, xfs_agnumber_t agno)
 		do_error("%s\n", strerror(error));
 	}
 	libxfs_sb_from_disk(sbp, buf);
-	libxfs_sb_quota_from_disk(sbp);
 
 	rval = verify_sb((char *)buf, sbp, agno == 0);
 	free(buf);
@@ -630,58 +657,13 @@ get_sb_geometry(fs_geometry_t *geo, xfs_sb_t *sbp)
 	if (xfs_sb_version_hasalign(sbp))
 		geo->sb_ialignbit = 1;
 
-	if (xfs_sb_version_hasshared(sbp) ||
-	    sbp->sb_versionnum & XR_PART_SECSB_VNMASK)
-		geo->sb_sharedbit = 1;
-
 	if (xfs_sb_version_hasdalign(sbp))
 		geo->sb_salignbit = 1;
 
 	if (xfs_sb_version_hasextflgbit(sbp))
 		geo->sb_extflgbit = 1;
 
-	/*
-	 * protect against pre-6.5 mkfs-generated garbaged
-	 * fields in the secondary superblocks.  pay attention
-	 * to those fields if and only if their corresponding
-	 * feature bits are set in the feature bits of the
-	 * version number or we can deduce from the version bits
-	 * that are set that our field was properly initialized
-	 * because a field after the field we care about was
-	 * properly initialized as well.
-	 */
-
-	/*
-	 * inode alignment field lives before the data alignment field
-	 */
-	if ((!pre_65_beta && (sbp->sb_versionnum & XR_PART_SECSB_VNMASK)) ||
-	    (pre_65_beta && (sbp->sb_versionnum & XR_ALPHA_SECSB_VNMASK)))
-		geo->sb_inoalignmt = sbp->sb_inoalignmt;
-
-	if ((!pre_65_beta && (sbp->sb_versionnum & XR_GOOD_SECSB_VNMASK)) ||
-	    (pre_65_beta && xfs_sb_version_hasdalign(sbp))) {
-		geo->sb_unit = sbp->sb_unit;
-		geo->sb_width = sbp->sb_width;
-	}
-
-	/*
-	 * shared vn always set if either ino or data alignment is on
-	 * since that field lives between the quota and inode alignment
-	 * fields
-	 */
-	if (sbp->sb_versionnum & XR_PART_SECSB_VNMASK)
-		geo->sb_shared_vn = sbp->sb_shared_vn;
-
-	/*
-	 * superblock fields located after sb_widthfields get set
-	 * into the geometry structure only if we can determine
-	 * from the features enabled in this superblock whether
-	 * or not the sector was zero'd at mkfs time.
-	 */
-	if ((!pre_65_beta && (sbp->sb_versionnum & XR_GOOD_SECSB_VNMASK)) ||
-	    (pre_65_beta && (sbp->sb_versionnum & XR_ALPHA_SECSB_VNMASK))) {
-		geo->sb_fully_zeroed = 1;
-	}
+	geo->sb_fully_zeroed = 1;
 }
 
 /*
@@ -702,20 +684,11 @@ verify_set_primary_sb(xfs_sb_t		*rsb,
 	xfs_sb_t	*sb;
 	fs_geo_list_t	*list;
 	fs_geo_list_t	*current;
-	char		*checked;
 	xfs_agnumber_t	agno;
 	int		num_sbs;
-	int		skip;
 	int		size;
 	int		num_ok;
 	int		retval;
-	int		round;
-
-	/*
-	 * select the number of secondaries to try for
-	 */
-	num_sbs = MIN(NUM_SBS, rsb->sb_agcount);
-	skip = howmany(num_sbs, rsb->sb_agcount);
 
 	/*
 	 * We haven't been able to validate the sector size yet properly
@@ -724,61 +697,48 @@ verify_set_primary_sb(xfs_sb_t		*rsb,
 	 * sector size rather than the sector size in @rsb.
 	 */
 	size = NUM_AGH_SECTS * (1 << (XFS_MAX_SECTORSIZE_LOG));
-	retval = 0;
 	list = NULL;
 	num_ok = 0;
 	*sb_modified = 0;
+	num_sbs = rsb->sb_agcount;
 
 	sb = (xfs_sb_t *) alloc_ag_buf(size);
-	checked = calloc(rsb->sb_agcount, sizeof(char));
-	if (!checked) {
-		do_error(_("calloc failed in verify_set_primary_sb\n"));
-		exit(1);
-	}
 
 	/*
 	 * put the primary sb geometry info onto the geometry list
 	 */
-	checked[sb_index] = 1;
 	get_sb_geometry(&geo, rsb);
 	list = add_geo(list, &geo, sb_index);
 
 	/*
-	 * grab N secondaries.  check them off as we get them
-	 * so we only process each one once
+	 * scan the secondaries and check them off as we get them so we only
+	 * process each one once
 	 */
-	for (round = 0; round < skip; round++)  {
-		for (agno = round; agno < rsb->sb_agcount; agno += skip)  {
-			if (checked[agno])
-				continue;
+	for (agno = 1; agno < rsb->sb_agcount; agno++) {
+		off = (xfs_off_t)agno * rsb->sb_agblocks << rsb->sb_blocklog;
 
-			off = (xfs_off_t)agno * rsb->sb_agblocks << rsb->sb_blocklog;
+		retval = get_sb(sb, off, size, agno);
+		if (retval == XR_EOF)
+			goto out_free_list;
 
-			checked[agno] = 1;
-			retval = get_sb(sb, off, size, agno);
-			if (retval == XR_EOF)
-				goto out_free_list;
-
-			if (retval == XR_OK) {
-				/*
-				 * save away geometry info.
-				 * don't bother checking the sb
-				 * against the agi/agf as the odds
-				 * of the sb being corrupted in a way
-				 * that it is internally consistent
-				 * but not consistent with the rest
-				 * of the filesystem is really really low.
-				 */
-				get_sb_geometry(&geo, sb);
-				list = add_geo(list, &geo, agno);
-				num_ok++;
-			}
+		if (retval == XR_OK) {
+			/*
+			 * save away geometry info. don't bother checking the
+			 * sb against the agi/agf as the odds of the sb being
+			 * corrupted in a way that it is internally consistent
+			 * but not consistent with the rest of the filesystem is
+			 * really really low.
+			 */
+			get_sb_geometry(&geo, sb);
+			list = add_geo(list, &geo, agno);
+			num_ok++;
 		}
 	}
 
 	/*
 	 * see if we have enough superblocks to bother with
 	 */
+	retval = 0;
 	if (num_ok < num_sbs / 2) {
 		retval = XR_INSUFF_SEC_SB;
 		goto out_free_list;
@@ -867,6 +827,5 @@ verify_set_primary_sb(xfs_sb_t		*rsb,
 out_free_list:
 	free_geo(list);
 	free(sb);
-	free(checked);
-	return(retval);
+	return retval;
 }
